#! /usr/bin/env python3
import ast
import os
import sys
import fnmatch
import ipaddress
from vpp_papi import macaddress
from vpp_papi import VPPApiClient as VPP
from vpp_papi import VppEnum

VPP_JSON_DIR = '/usr/share/vpp/api/'
API_FILE_SUFFIX = '*.api.json'

"""
brief An enumeration of the type of ports that can be added to a bridge domain
  /* a 'normal' interface, i.e. not BVI or UU-Flood */
  L2_API_PORT_TYPE_NORMAL = 0,
  /* a BVI interface in the BD */
  L2_API_PORT_TYPE_BVI = 1,
  /* The interface on which to forward unknown unicast packets
   * If this is not set for a BD then UU is flooded */
  L2_API_PORT_TYPE_UU_FWD = 2,

"""
L2_API_PORT_TYPE_NORMAL = 0
L2_API_PORT_TYPE_BVI = 1
L2_API_PORT_TYPE_UU_FWD = 2

"""
brief Flags that can be changed on a bridge domain
"""
BRIDGE_API_FLAG_NONE = 0x0
BRIDGE_API_FLAG_LEARN = 0x1
BRIDGE_API_FLAG_FWD = 0x2
BRIDGE_API_FLAG_FLOOD = 0x4
BRIDGE_API_FLAG_UU_FLOOD = 0x8
BRIDGE_API_FLAG_ARP_TERM = 0x10
BRIDGE_API_FLAG_ARP_UFWD = 0x20

ROUTE_TYPE_NORMAL = 0
ROUTE_TYPE_LOCAL = 1
PROTO_IPV4 = 0
PROTO_IPV6 = 1

L2_VTR_DISABLED = 0
L2_VTR_PUSH_1 = 1
L2_VTR_PUSH_2 = 2
L2_VTR_POP_1 = 3
L2_VTR_POP_2 = 4
L2_VTR_TRANSLATE_1_1 = 5
L2_VTR_TRANSLATE_1_2 = 6
L2_VTR_TRANSLATE_2_1 = 7
L2_VTR_TRANSLATE_2_2 = 8

IP_API_NEIGHBOR_FLAG_NONE = 0
IP_API_NEIGHBOR_FLAG_STATIC = 1
IP_API_NEIGHBOR_FLAG_NO_FIB_ENTRY = 2

AF_XDP_API_MODE_AUTO = 0
AF_XDP_API_MODE_COPY = 1
AF_XDP_API_MODE_ZERO_COPY = 2


AF_XDP_API_FLAGS_SYSCALL_DEFAULT = 0
AF_XDP_API_FLAGS_NO_SYSCALL_LOCK = 1


pppoe_states = dict([(0, 'PPPOE_CLIENT_DISCOVERY'),
                    (1, 'PPPOE_CLIENT_REQUEST'), (2, 'PPPOE_CLIENT_SESSION')])


def mac_to_bytes(hw_addr):
    if not hw_addr:
        return None
    return bytes.fromhex(hw_addr.replace(':', ''))


def binary_type(s):
    return s.encode('ascii')


def fix_string(s):
    if isinstance(s, str):
        return(s)
    return s.decode('ascii').rstrip('\0')


def bytes_to_ip(ip_bytes, is_ipv6):
    if is_ipv6:
        return str(ipaddress.ip_address(ip_bytes))
    else:
        return str(ipaddress.ip_address(ip_bytes[:4]))


class VppApiInterface(object):

    def __init__(self, client_name, read_timeout=None, rx_qlen=512):
        # self.LOG = log
        jsonfiles = []
        for root, dirnames, filenames in os.walk(VPP_JSON_DIR):
            for filename in fnmatch.filter(filenames, pat=API_FILE_SUFFIX):
                jsonfiles.append(os.path.join(root, filename))

        if not jsonfiles:
            print('Error: no json api files found')
            exit(0)

        args = {}
        if read_timeout is not None:
            args['read_timeout'] = read_timeout
        args['rx_qlen'] = rx_qlen
        self._vpp = VPP(apifiles=jsonfiles)
        self._vpp.connect(client_name, **args)

    def __del__(self):
        self._vpp.disconnect()

    def get_version(self):
        r = self._vpp.api.show_version()
        if r.retval == 0:
            return r.version
        return

    # ##### BRIDGE DOMAINS #####
    """ brief L2 bridge domain add or delete request
		@param client_index - opaque cookie to identify the sender
		@param context - sender context, to match reply w/ request
		@param bd_id - the bridge domain to create
		@param flood - enable/disable bcast/mcast flooding in the bd
		@param uu_flood - enable/disable unknown unicast flood in the bd
		@param forward - enable/disable forwarding on all interfaces in the bd
		@param learn - enable/disable learning on all interfaces in the bd
		@param arp_term - enable/disable arp termination in the bd
		@param arp_ufwd - enable/disable arp unicast forwarding in the bd
		@param mac_age - mac aging time in min, 0 for disabled
		@param is_add - add or delete flag
	"""

    def create_bridge_domain(self, bd_id, flood=True, uu_flood=True, forward=True,
                             learn=True, arp_term=False, arp_ufwd=False, mac_age=0, tag=""):
        ret = self._vpp.api.bridge_domain_add_del_v2(
            is_add=True,
            bd_id=bd_id,
            flood=flood,
            uu_flood=uu_flood,
            forward=forward,
            learn=learn,
            arp_term=arp_term,
            arp_ufwd=arp_ufwd,
            bd_tag=tag,
            mac_age=mac_age)
        return ret.retval

    def delete_bridge_domain(self, bd_id):
        ret = self._vpp.api.bridge_domain_add_del_v2(
            is_add=False,
            bd_id=bd_id)
        return ret.retval

    def get_bridge_domains_idxs(self):
        ret = self._vpp.api.bridge_domain_dump(bd_id=0xffffffff)
        return set([bd.bd_id for bd in ret])

    """brief L2 bridge domain operational state response
		@param bd_id - the bridge domain id
		@param flood - bcast/mcast flooding state on all interfaces in the bd
		@param uu_flood - unknown unicast flooding state on all interfaces in the bd
		@param forward - forwarding state on all interfaces in the bd
		@param learn - learning state on all interfaces in the bd
		@param arp_term - arp termination state on all interfaces in the bd
		@param arp_ufwd - arp unicast forwarding state on all interfaces in the bd
		@param mac_age - mac aging time in min, 0 for disabled
		@param bd_tag - optional textual tag for the bridge domain
		@param n_sw_ifs - number of sw_if_index's in the domain
		{
		u32 context;
		u32 bd_id;
		bool flood;
		bool uu_flood;
		bool forward;
		bool learn;
		bool arp_term;
		bool arp_ufwd;
		u8 mac_age;
		string bd_tag[64];
		vl_api_interface_index_t bvi_sw_if_index;
		vl_api_interface_index_t uu_fwd_sw_if_index;
		u32 n_sw_ifs;
		vl_api_bridge_domain_sw_if_t sw_if_details[n_sw_ifs];
		};
	"""

    def get_bridge_domains(self, bd_id=0xffffffff):
        ret = self._vpp.api.bridge_domain_dump(bd_id=bd_id)
        for bd in ret:
            if_details = []
            for intf in bd.sw_if_details:
                if_details.append(
                    {'sw_if_index': intf.sw_if_index, 'shg': intf.shg})

            yield {'bd_id': bd.bd_id,
                   'tag': fix_string(bd.bd_tag),
                   'mac_age': bd.mac_age,
                   'flood': bd.flood,
                   'uu_flood': bd.uu_flood,
                   'forward': bd.forward,
                   'learn': bd.learn,
                   'arp_term': bd.arp_term,
                   'arp_ufwd': bd.arp_ufwd,
                   'bvi_sw_if_index': bd.bvi_sw_if_index,
                   'uu_fwd_sw_if_index': bd.uu_fwd_sw_if_index,
                   'n_sw_ifs': bd.n_sw_ifs,
                   'sw_if_details': if_details
                   }

    def get_bridge_domain(self, bd_id):
        for bd in self.get_bridge_domains():
            if bd_id == bd['bd_id']:
                return bd
        return None

    def get_bd_bvi_by_bdidx(self, bd_id):
        bd = self.get_bridge_domain(bd_id=bd_id)
        if bd != None:
            return bd['bvi_sw_if_index']
        return None

    def get_bd_by_tag(self, tag):
        for bd in self.get_bridge_domains():
            if tag == bd['tag']:
                return bd
        return None

    def get_bd_id_by_tag(self, tag):
        for bd in self.get_bridge_domains():
            if tag == bd['tag']:
                return bd['bd_id']
        return None

    """ brief Set bridge flags request
		@param client_index - opaque cookie to identify the sender
		@param context - sender context, to match reply w/ request
		@param bd_id - the bridge domain to set the flags for
		@param is_set - if non-zero, set the flags, else clear them
		@param flags - flags that are non-zero to set or clear
	"""

    def set_bd_flags(self, bd_id, is_set, flags):
        ret = self._vpp.api.bridge_flags(
            bd_id=bd_id, is_set=is_set, flags=flags)
        return ret.retval

    def set_bd_arp_term(self, bd_id):
        return self.set_bd_flags(bd_id=bd_id, is_set=1, flags=BRIDGE_API_FLAG_ARP_TERM)

    def unset_bd_arp_term(self, bd_id):
        return self.set_bd_flags(bd_id=bd_id, is_set=0, flags=BRIDGE_API_FLAG_ARP_TERM)

    # ----- BRIDGE DOMAIN ARP ENTRIES -----
    """ brief Set bridge domain ip to mac entry request
		@param client_index - opaque cookie to identify the sender
		@param context - sender context, to match reply w/ request
		@param bd_id - the bridge domain to set the flags for
		@param is_add - if non-zero, add the entry, else clear it
		@param ip - ipv4 or ipv6 address
		@param mac - MAC address
	"""

    def set_bd_ip_mac_entry(self, is_add, entry):
        ret = self._vpp.api.bd_ip_mac_add_del(is_add=is_add, entry=entry)
        return ret.retval

    def add_bd_ip_mac_entry(self, bd_id, ip=None, mac=None):
        if ip is None or mac is None:
            return -1

        return self.set_bd_ip_mac_entry(
            is_add=1,
            entry={'bd_id': bd_id,
                   'ip': ipaddress.ip_address(ip),
                   'mac': mac
                   })

    def del_bd_ip_mac_entry(self, bd_id, ip=None, mac=None):
        if ip is None or mac is None:
            return -1

        return self.set_bd_ip_mac_entry(
            is_add=0,
            entry={'bd_id': bd_id,
                   'ip': ipaddress.ip_address(ip),
                   'mac': mac
                   })

    """ brief Flush bridge domain IP to MAC entries
		@param client_index - opaque cookie to identify the sender
		@param bd_id - bridge domain identifier
	"""

    def flush_bd_ip_mac_entries(self, bd_id):
        ret = self._vpp.api.bd_ip_mac_flush(bd_id=bd_id)
        return ret.retval

    """ brief bridge domain IP to MAC entry details structure
		@param bd_id - bridge domain table id
		--no-- @param is_ipv6 - if non-zero, ipv6 address, else ipv4 address
		@param ip_address - ipv4 or ipv6 address
		@param mac_address - MAC address
	"""

    def get_bd_ip_mac_entries(self, bd_id):
        ret = self._vpp.api.bd_ip_mac_dump(bd_id=bd_id)
        for entry in ret:
            yield {'bd_id': entry.entry.bd_id,
                   'ip_addr': str(entry.entry.ip),
                   'mac_addr': str(entry.entry.mac)
                   }

    def get_bd_ip_mac_entry_by_ip(self, bd_id, ip):
        ret = self.get_bd_ip_mac_entries(bd_id)
        for entry in ret:
            if ip == entry['ip_addr']:
                return entry
        return None

    def get_bd_ip_mac_entry_by_mac(self, bd_id, mac):
        ret = self.get_bd_ip_mac_entries(bd_id)
        for entry in ret:
            if mac == entry['mac_addr']:
                return entry
        return None

    # ##### INTERFACES #####
    def search_vppif_by_hostifname(self, hostifname):
        """ Find vpp interface by hostifname.
                Search through all the different type of interfaces:
                        - dpdk interfaces:	vppname == hostifname
                        - host interface:	vppname == 'host-' + hostifname
                        - tap interfaces:	vppname[host_if_name'] == hostifname
                        - vlan subinterface:
                                                dpdk:	vppname == hostifname.vlan
                                        virtio:	vppname == 'host-' + hostifname.vlan
        """
        intfc = None

        if 'chain' in hostifname:
            hostifname = 'br-' + hostifname

        intfc = self.get_if_by_name(ifname=hostifname)
        if intfc != None:
            return (intfc['ifname'], intfc['sw_if_index'])

        # HOST INTERFACES
        intfc = self.get_host_interface(ifname=hostifname)
        if intfc != None:
            return ('host-'+intfc['host_if_name'], intfc['sw_if_index'])

        # TAP INTERFACES
        intfc = self.get_tap_interface_by_host_ifname(ifname=hostifname)
        if intfc != None:
            return (intfc['ifname'], intfc['sw_if_index'])

        # VLAN SUB INTERFACE DPDK
        intfc = self.get_vlan_sub_interface(ifname=hostifname)
        if intfc != None:
            return (intfc['ifname'], intfc['sw_if_index'])

        # VLAN SUB INTERFACE DPDK
        intfc = self.get_vlan_sub_interface(ifname='host-'+hostifname)
        if intfc != None:
            return (intfc['ifname'], intfc['sw_if_index'])

        return (None, None)

    def get_interfaces(self):
        """	brief Request all or filtered subset of sw_interface_details
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - index of the interface to dump info on, 0 or ~0 if on all
                TODO: Support selecting only index==0 when CSIT is ready.
                @param name_filter_valid - 1 if requesting a filtered subset of records else 0
                if name filter is set as valid, sw_if_index value is ignored and all interfaces are examined
                @param name_filter - interface name substring filter. Eg. loop1 returns [loop1, loop10]

                define sw_interface_details
                        {
                        u32 context;
                        vl_api_interface_index_t sw_if_index;

                        /* index of sup interface (e.g. hw interface).
                                equal to sw_if_index for super hw interface. */
                        u32 sup_sw_if_index;

                        /* Layer 2 address, if applicable */
                        vl_api_mac_address_t l2_address;

                        vl_api_if_status_flags_t flags;

                        vl_api_if_type_t type;

                        /* 1 = half duplex, 2 = full duplex */
                        vl_api_link_duplex_t link_duplex;

                        /* link speed in kbps */
                        u32 link_speed;

                        /* MTU */
                        u16 link_mtu;

                        /* Per protocol MTUs */
                        u32 mtu[4]; /* vl_api_mtu_proto_t 0 - L3, 1 - IP4, 2 - IP6, 3 - MPLS */

                        /* Subinterface ID. A number 0-N to uniquely identify this subinterface under the super interface */
                        u32 sub_id;

                        /* Number of tags 0-2 */
                        u8 sub_number_of_tags;
                        u16 sub_outer_vlan_id;
                        u16 sub_inner_vlan_id;

                        vl_api_sub_if_flags_t sub_if_flags;

                        /* vlan tag rewrite state */
                        u32 vtr_op;
                        u32 vtr_push_dot1q;		// ethertype of first pushed tag is dot1q/dot1ad
                        u32 vtr_tag1;			// first pushed tag
                        u32 vtr_tag2;			// second pushed tag

                        /* pbb tag rewrite info */
                        u16 outer_tag;
                        vl_api_mac_address_t  b_dmac;
                        vl_api_mac_address_t  b_smac;
                        u16 b_vlanid;
                        u32 i_sid;

                        /* Interface name */
                        string interface_name[64];
                        string interface_dev_type[64];
                        string tag[64];
                        };

        """
        ret = self._vpp.api.sw_interface_dump()
        for iface in ret:
            # mac = bytearray(iface.l2_address[:iface.l2_address_length])
            yield {'ifname': fix_string(iface.interface_name),
                   'tag': fix_string(iface.tag),
                   'mac_addr': iface.l2_address,
                   'sw_if_index': iface.sw_if_index,
                   'sup_sw_if_index': iface.sup_sw_if_index,
                   # DI: flags should be devided into admin_up_down and link_up_down
                   'admin_up_down': iface.flags,
                   'link_up_down': iface.flags,
                   'sub_id': iface.sub_id,
                   'interface_dev_type': fix_string(iface.interface_dev_type),
                   }

    def get_if_by_name(self, ifname):
        for iface in self.get_interfaces():
            if iface['ifname'] == ifname:
                return iface
        return None

    def get_if_by_ifidx(self, idx):
        for iface in self.get_interfaces():
            if iface['sw_if_index'] == idx:
                return iface
        return None

    def get_ifname_by_ifidx(self, idx):
        for iface in self.get_interfaces():
            if iface['sw_if_index'] == idx:
                return iface['ifname']
        return None

    def get_ifidx_by_name(self, ifname):
        for iface in self.get_interfaces():
            if iface['ifname'] == ifname:
                return iface['sw_if_index']
        return None

    """ brief Set flags on the interface
		@param client_index - opaque cookie to identify the sender
		@param context - sender context, to match reply w/ request
		@param sw_if_index - index of the interface to set flags on
		@param flags - interface_status flags
			(only IF_STATUS_API_FLAG_ADMIN_UP used in config)
	"""

    def ifup(self, idx):
        ret = self._vpp.api.sw_interface_set_flags(sw_if_index=idx, flags=1)
        return ret.retval

    def ifdown(self, idx):
        ret = self._vpp.api.sw_interface_set_flags(sw_if_index=idx, flags=0)
        return ret.retval

    def register_event_callback(self, callback):
        """ brief Register async callback
                @param callback func
        """
        return self._vpp.register_event_callback(callback)

    def want_interface_events(self, enable_disable):
        """ brief Register for interface events
                @param enable_disable - 1 => register for events, 0 => cancel registration
                @param pid - sender's pid
        """
        pid = os.getpid()
        return self._vpp.api.want_interface_events(enable_disable=enable_disable, pid=pid)

    def add_to_bridge(self, bd_id, idx, port_type=L2_API_PORT_TYPE_NORMAL, shg=0):
        """	brief Interface bridge mode request
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param rx_sw_if_index - the interface
                @param bd_id - bridge domain id
                @param port_type - port_mode, see #l2_port_type
                @param shg - Split horizon group, for bridge mode only
                @param enable - Enable beige mode if not 0, else set to L3 mode
        """
        ret = self._vpp.api.sw_interface_set_l2_bridge(
            rx_sw_if_index=idx,
            bd_id=bd_id,
            port_type=port_type,
            shg=shg,              # shared horizon group
            enable=True)        # enable bridge mode
        return ret.retval

    def del_from_bridge(self, bd_id, idx, port_type=L2_API_PORT_TYPE_NORMAL):
        ret = self._vpp.api.sw_interface_set_l2_bridge(
            rx_sw_if_index=idx,
            bd_id=bd_id,
            port_type=port_type,
            shg=0,              # shared horizon group
            enable=False)        # disable bridge mode
        return ret.retval

    def check_interface_in_bridge(self, bd_id, idx):
        bd = self.get_bridge_domain(bd_id=bd_id)
        if bd is None or 0 == bd.get('n_sw_ifs', 0):
            return False

        for intf in bd.get('sw_if_details'):
            if idx == intf['sw_if_index']:
                return True
        return False

    def get_interface_bd(self, idx):
        bds = self.get_bridge_domains()
        for bd in bds:
            bd_ifs = list(map(lambda x: x['sw_if_index'], bd['sw_if_details']))
            if idx in bd_ifs:
                return bd
        return None

    def get_interface_bd_id(self, idx):
        bds = self.get_bridge_domains()
        for bd in bds:
            bd_ifs = list(map(lambda x: x['sw_if_index'], bd['sw_if_details']))
            if idx in bd_ifs:
                return bd['bd_id']
        return None

    def __get_rx_mode(self, mode_id):

        mode = 'unknown'

        if mode_id == 1:
            mode = 'polling'
        elif mode_id == 2:
            mode = 'interrupt'
        elif mode_id == 3:
            mode = 'adaptive'

        return mode

    def __set_rx_mode(self, mode):

        mode_id = 0

        if mode == 'polling' or mode == 1:
            mode_id = 1
        elif mode == 'interrupt' or mode == 2:
            mode_id = 2
        elif mode == 'adaptive' or mode == 3:
            mode_id = 3

        return mode_id

    def set_interface_rx_mode(self, idx, mode, queue_id_valid=0, queue_id=0):
        """ brief Set an interface's rx-mode
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - the interface whose rx-mode will be set
                @param queue_id_valid - 1 = the queue_id field is valid. 0 means all
                        queue_id's
                @param queue_id - the queue number whose rx-mode will be set. Only valid
                        if queue_id_valid is 1
                @param mode - polling=1, interrupt=2, adaptive=3
        """
        rx_mode = self.__set_rx_mode(mode)
        if rx_mode == 0:
            print("Unknown RX interface mode")
            return 1
        ret = self._vpp.api.sw_interface_set_rx_mode(
            sw_if_index=idx, mode=rx_mode, queue_id_valid=queue_id_valid, queue_id=queue_id)
        return ret.retval

    def set_interface_learning(self, idx, disable=0):
        """brief Set interface L2 flags (such as L2_LEARN, L2_FWD, 
                        L2_FLOOD, L2_UU_FLOOD, or L2_ARP_TERM bits). This can be used
                        to disable one or more of the features represented by the
                        flag bits on an interface to override what is set as default
                        for all interfaces in the bridge domain
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - interface 
                @param is_set - if non-zero, set the bits, else clear them
                @param feature_bitmap - non-zero bits (as above) to set or clear
        """
        flag_l2_learn = (1 << 0)
        ret = self._vpp.api.l2_flags(
            sw_if_index=idx, feature_bitmap=flag_l2_learn, is_set=(1 if disable == 0 else 0))
        return ret.retval

    def set_interface_mtu(self, idx, mtu):
        """ MTU for L3, IPv4, IPv6 and MPLS:
                u32 mtu[4]; 
                Where 0 - L3, 1 - IP4, 2 - IP6, 3 - MPLS
        """
        l3_mtu = [mtu, 0, 0, 0]
        ret = self._vpp.api.sw_interface_set_mtu(sw_if_index=idx, mtu=l3_mtu)
        return ret.retval

    def get_interfaces_rx_placement(self, idx=0xffffffff):
        """brief dump the rx queue placement of interface(s)
                @param sw_if_index - optional interface index for which queue placement to
                        be requested. sw_if_index = ~0 will dump placement information for all
                        interfaces. It will not dump information related to sub-interfaces, p2p
                        and pipe interfaces.

                show the interface's queue - thread placement
                        This api is used to display the interface and queue worker
                        thread placement. One message per rx-queue per interface will
                        be sent to client.
                        Each message will contain information about rx-queue id of an
                        interface, interface index, thread on which this rx-queue is
                        placed and mode of rx-queue.
                        @param client_index - opaque cookie to identify the sender
                        @param context - sender context, to match reply w/ request
                        @param sw_if_index - the interface whose rx-placement will be dumped
                        @param queue_id - the queue id
                        @param worker_id - the worker id on which queue_id is placed,
                                                        worker_id = 0 means main thread.
                        @param mode - polling=1, interrupt=2, adaptive=3
        """
        ret = self._vpp.api.sw_interface_rx_placement_dump(sw_if_index=idx)
        for iface in ret:
            yield {'sw_if_index': iface.sw_if_index,
                   'queue_id': iface.queue_id,
                   'worker_id': iface.worker_id,
                   'rx-mode': self.__get_rx_mode(iface.mode)}

    def get_interface_rx_placement(self, idx):
        for iface in self.get_interfaces_rx_placement():
            if iface['sw_if_index'] == idx:
                return iface
        return None

    def add_interface_address(self, idx, addr):
        """brief Set or delete one or all ip addresses on a specified interface
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - index of the interface to add/del addresses
                @param is_add - add address if non-zero, else delete
                @param del_all - if non-zero delete all addresses on the interface
                @param prefix - address + a prefix length for the implied connected route
        """
        ipaddr = ipaddress.ip_interface(addr)
        ret = self._vpp.api.sw_interface_add_del_address(
            sw_if_index=idx,
            prefix=ipaddr,
            is_add=1, del_all=0)
        return ret.retval

    def del_interface_address(self, idx, addr):
        """brief Set or delete one or all ip addresses on a specified interface
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - index of the interface to add/del addresses
                @param is_add - add address if non-zero, else delete
                @param del_all - if non-zero delete all addresses on the interface
                @param prefix - address + a prefix length for the implied connected route
        """
        ipaddr = ipaddress.ip_interface(addr)
        ret = self._vpp.api.sw_interface_add_del_address(
            sw_if_index=idx,
            prefix=ipaddr,
            is_add=0, del_all=0)
        return ret.retval

    def del_all_interface_addresses(self, idx):
        """brief Set or delete one or all ip addresses on a specified interface
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - index of the interface to add/del addresses
                @param is_add - add address if non-zero, else delete
                @param del_all - if non-zero delete all addresses on the interface
                @param prefix - address + a prefix length for the implied connected route
        """
        ret = self._vpp.api.sw_interface_add_del_address(
            sw_if_index=idx, is_add=0, del_all=1)
        return ret.retval

    def get_interface_address_by_ifidx(self, ifidx):
        ret = self._vpp.api.ip_address_dump(
            sw_if_index=ifidx
        )
        for ip in ret:
            return ip.prefix

    def set_interface_vrf(self, idx, table_id, is_ipv6=False):
        """brief Associate the specified interface with a fib table
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - index of the interface
                @param is_ipv6 - if non-zero ipv6, else ipv4
                @param vrf_id - fib table/vrf id to associate the interface with
        """
        ret = self._vpp.api.sw_interface_set_table(
            sw_if_index=idx, vrf_id=table_id, is_ipv6=is_ipv6)
        return ret.retval

    def get_interface_vrf(self, idx):
        """brief Get VRF id assigned to interface
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - index of the interface

                brief Reply to get_sw_interface_vrf
            @param vrf_id - VRF id assigned to the interface
        """
        ret = self._vpp.api.sw_interface_get_table(sw_if_index=idx)
        return (ret.retval, ret.vrf_id)

    # ##### SUB INTERFACES #####
    def create_vlan_subinterface(self, idx, vlan_id):
        """brief Create a new subinterface with the given vlan id
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - software index of the new vlan's parent interface
                @param vlan_id - vlan tag of the new interface
        """
        ret = self._vpp.api.create_vlan_subif(sw_if_index=idx, vlan_id=vlan_id)
        return (ret.retval, ret.sw_if_index)

    def delete_vlan_subinterface(self, idx):
        """brief Delete sub interface request
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - sw index of the interface that was created by create_subif
        """
        ret = self._vpp.api.delete_subif(sw_if_index=idx)
        return ret.retval

    def get_vlan_sub_interfaces(self):
        ifaces = self.get_interfaces()
        for iface in ifaces:
            if iface['sub_id'] > 0:
                yield {'ifname': fix_string(iface['ifname']),
                       'sw_if_index': iface['sw_if_index'],
                       'sub_id': iface['sub_id']}

    def get_vlan_sub_interface(self, ifname):
        for iface in self.get_vlan_sub_interfaces():
            if iface['ifname'] == ifname:
                return iface
        return None

    def is_vlan_sub_interface_by_idx(self, idx):
        for iface in self.get_vlan_sub_interfaces():
            if iface['sw_if_index'] == idx:
                return iface
        return None

    def is_vlan_sub_interface_idx_by_name(self, ifname):
        for iface in self.get_vlan_sub_interfaces():
            if iface['ifname'] == ifname:
                return iface['sw_if_index']
        return None

    def is_vlan_sub_interface(self, ifname):
        for iface in self.get_vlan_sub_interfaces():
            if iface['ifname'] == ifname:
                return True
        return False

    def vlan_rewrite_disable(self, idx):
        """brief L2 interface vlan tag rewrite configure request
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - interface the operation is applied to
                @param vtr_op - Choose from l2_vtr_op_t enum values
                @param push_dot1q - first pushed flag dot1q id set, else dot1ad
                @param tag1 - Needed for any push or translate vtr op
                @param tag2 - Needed for any push 2 or translate x-2 vtr ops
        """
        ret = self._vpp.api.l2_interface_vlan_tag_rewrite(
            sw_if_index=idx, vtr_op=L2_VTR_DISABLED, push_dot1q=0, tag1=0, tag2=0)
        return ret.retval

    def vlan_rewrite_poptag1(self, idx):
        ret = self._vpp.api.l2_interface_vlan_tag_rewrite(
            sw_if_index=idx, vtr_op=L2_VTR_POP_1, push_dot1q=0, tag1=0, tag2=0)
        return ret.retval

    def vlan_rewrite_poptag2(self, idx):
        ret = self._vpp.api.l2_interface_vlan_tag_rewrite(
            sw_if_index=idx, vtr_op=L2_VTR_POP_2, push_dot1q=0, tag1=0, tag2=0)
        return ret.retval

    # ##### HOST INTERFACES #####

    def create_host_interface(self, hostifname, mac_addr=None, use_random_hw_addr=True, packet_mode=1):
        """ brief Create host-interface
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param host_if_name - interface name
                @param hw_addr - interface MAC
                @param use_random_hw_addr - use random generated MAC
        """
        if mac_addr is not None:
            hw_addr = mac_addr
            use_random_hw_addr = False
        else:
            hw_addr = '00:00:00:00:00:00'

        ret = self._vpp.api.af_packet_create_v3(
            host_if_name=hostifname,
            hw_addr=hw_addr,
            use_random_hw_addr=use_random_hw_addr,
            mode=packet_mode)

        return (ret.retval, ret.sw_if_index)

    def delete_host_interface(self, hostifname):
        """ brief Delete host-interface
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param host_if_name - interface name
        """
        ret = self._vpp.api.af_packet_delete(host_if_name=hostifname)
        return ret.retval

    def get_host_interfaces(self):
        ret = self._vpp.api.af_packet_dump()
        for iface in ret:
            yield {'host_if_name': fix_string(iface.host_if_name),
                   'sw_if_index': iface.sw_if_index}

    def get_host_interface(self, ifname):
        for iface in self.get_host_interfaces():
            if iface['host_if_name'] == ifname:
                return iface
        return None

    def is_host_interface(self, hostifname):
        for iface in self.get_host_interfaces():
            if iface['host_if_name'] == hostifname:
                return True
        return False

    def is_host_interface_by_ifidx(self, idx):
        for iface in self.get_host_interfaces():
            if iface['sw_if_index'] == idx:
                return True
        return False

    def get_host_interface_by_ifidx(self, idx):
        for iface in self.get_host_interfaces():
            if iface['sw_if_index'] == idx:
                return iface
        return None

    # ##### TAP INTERFACES #####

    def create_tap_interface(self, hostifname, id=0xffffffff, mac_addr=None, use_random_hw_addr=True,
                             host_ip4_addr_set=False, host_bridge_set=False,
                             host_namespace_set=False, host_namespace=None,
                             host_mac_addr_set=False, host_mac_addr=None, host_ip4_addr=None,
                             host_ip6_addr_set=False, host_ip6_addr=None,
                             tag="", host_mtu_size=None):
        """ brief Initialize a new tap interface with the given parameters
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param id - interface id, 0xffffffff means auto
                @param use_random_mac - let the system generate a unique mac address
                @param mac_address - mac addr to assign to the interface if use_random not set
                @param num_rx_queues - number of rx queues
                @param tx_ring_sz - the number of entries of TX ring, optional, default is 256 entries, must be power of 2
                @param rx_ring_sz - the number of entries of RX ring, optional, default is 256 entries, must be power of 2
                @param host_mtu_set - host MTU should be set
                @param host_mtu_size - host MTU size
                @param host_mac_addr_set - host side interface mac address should be set
                @param host_mac_addr - host side interface mac address
                @param host_ip4_prefix_set - host IPv4 ip address should be set
                @param host_ip4_prefix - host IPv4 ip address
                @param host_ip6_prefix_set - host IPv6 ip address should be set
                @param host_ip6_prefix - host IPv6 ip address
                @param host_ip4_gw_set - host IPv4 default gateway should be set
                @param host_ip4_gw - host IPv4 default gateway
                @param host_ip6_gw_set - host IPv6 default gateway should be set
                @param host_ip6_gw - host IPv6 default gateway
                @param tap_flags - flags for the TAP interface creation
                @param host_if_name_set - host side interface name should be set
                @param host_if_name - host side interface name
                @param host_namespace_set - host namespace should be set
                @param host_namespace - host namespace to attach interface to
                @param host_bridge_set - host bridge should be set
                @param host_bridge - host bridge to attach interface to
                @param tag - tag
        """
        if mac_addr is not None:
            hw_addr = mac_to_bytes(mac_addr)
            use_random_hw_addr = False
        else:
            hw_addr = mac_to_bytes('00:00:00:00:00:00')

        if host_mac_addr is not None:
            host_hw_addr = mac_to_bytes(host_mac_addr)
            host_mac_addr_set = True
        else:
            host_hw_addr = mac_to_bytes('00:00:00:00:00:00')

        host_mtu_set = False
        if host_mtu_size is not None:
            host_mtu_set = True

        ret = self._vpp.api.tap_create_v2(
            use_random_mac=use_random_hw_addr,
            mac_address=hw_addr,
            host_if_name_set=True,
            host_if_name=hostifname,
            id=id,
            host_ip4_prefix_set=host_ip4_addr_set,
            host_ip4_prefix=host_ip4_addr,
            host_bridge_set=host_bridge_set,
            host_namespace_set=host_namespace_set,
            host_namespace=host_namespace if host_namespace_set == True else None,
            host_mac_addr_set=host_mac_addr_set,
            host_mac_addr=host_hw_addr,
            host_ip6_prefix_set=host_ip6_addr_set,
            host_ip6_prefix=host_ip6_addr,
            host_mtu_set=host_mtu_set,
            host_mtu_size=host_mtu_size,
            rx_ring_sz=4096, tx_ring_sz=4096,
            tag=tag)

        return (ret.retval, ret.sw_if_index)

    def delete_tap_interface(self, idx):
        ret = self._vpp.api.tap_delete_v2(sw_if_index=idx)
        return ret.retval

    def get_tap_interfaces(self):
        """brief Reply for tap dump request
                @param sw_if_index - software index of tap interface
                @param id - interface id
                @param dev_name - Linux tap device name
                @param tx_ring_sz - the number of entries of TX ring
                @param rx_ring_sz - the number of entries of RX ring
                @param host_mtu_size - host mtu size
                @param host_mac_addr - mac address assigned to the host side of the interface
                @param host_ip4_prefix - host IPv4 ip address
                @param host_ip6_prefix - host IPv6 ip address
                @param tap_flags - flags for the TAP interface creation
                @param host_if_name - host side interface name
                @param host_namespace - host namespace the interface is attached into
                @param host_bridge - host bridge the interface is attached into
        """
        ret = self._vpp.api.sw_interface_tap_v2_dump()
        for iface in ret:
            yield {'ifname': fix_string(iface.dev_name),
                   'id': iface.id,
                   'sw_if_index': iface.sw_if_index,
                   'host_if_name': fix_string(iface.host_if_name),
                   'host_ns': fix_string(iface.host_namespace),
                   'host_mac_addr': iface.host_mac_addr,
                   'host_mtu_size': iface.host_mtu_size,
                   'host_ip4_prefix': iface.host_ip4_prefix,
                   'host_ip6_prefix': iface.host_ip6_prefix}

    def get_tap_interface(self, ifname):
        for iface in self.get_tap_interfaces():
            if iface['ifname'] == ifname:
                return iface
        return None

    def get_tap_interface_by_id(self, id):
        for iface in self.get_tap_interfaces():
            if iface['id'] == id:
                return iface
        return None

    def get_tap_interface_by_host_ifname(self, ifname):
        for iface in self.get_tap_interfaces():
            if iface['host_if_name'] == ifname:
                return iface
        return None

    def is_tap_interface(self, ifname):
        for iface in self.get_tap_interfaces():
            if iface['ifname'] == ifname:
                return True
        return False

    def is_tap_interface_by_ifidx(self, idx):
        for iface in self.get_tap_interfaces():
            if iface['sw_if_index'] == idx:
                return True
        return False

    # ##### LOOPBACK INTERFACES #####

    def create_loopback_interface(self, id=0xffffffff, mac_addr=None):
        """brief Create loopback interface instance request
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param mac_address - mac addr to assign to the interface if none-zero
                @param is_specified - if non-0, a specific user_instance is being requested
                @param user_instance - requested instance, ~0 => dynamically allocate
        """
        if mac_addr is not None:
            hw_addr = mac_to_bytes(mac_addr)
        else:
            hw_addr = mac_to_bytes('00:00:00:00:00:00')

        ret = self._vpp.api.create_loopback_instance(user_instance=id, is_specified=(
            1 if id != 0xffffffff else 0), mac_address=hw_addr)
        return (ret.retval, ret.sw_if_index)

    def delete_loopback_interface(self, idx):
        """brief Delete loopback interface request
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - sw index of the interface that was created
        """
        ret = self._vpp.api.delete_loopback(sw_if_index=idx)
        return ret.retval

    # ##### VHOSTUSER INTERFACES #####
    def create_vhostuser_interface(self, sock_filename, is_server=False):
        ret = self._vpp.api.create_vhost_user_if(
            is_server=is_server,
            sock_filename=sock_filename)
        return ret.retval

    def get_vhostuser_interfaces(self):
        """brief Vhost-user interface details structure (fix this)
                @param sw_if_index - index of the interface
                @param interface_name - name of interface
                @param virtio_net_hdr_sz - net header size
                @param features_first_32 - interface features, first 32 bits
                @param features_last_32 - interface features, last 32 bits
                @param is_server - vhost-user server socket
                @param sock_filename - socket filename
                @param num_regions - number of used memory regions
                @param sock_errno - socket errno
        """
        ret = self._vpp.api.sw_interface_vhost_user_dump()
        for iface in ret:
            yield {'ifname': fix_string(iface.interface_name),
                   'sw_if_index': iface.sw_if_index,
                   'is_server': iface.is_server,
                   'sock_filename': fix_string(iface.sock_filename),
                   'num_regions': iface.num_regions,
                   'sock_errno': iface.sock_errno}

    def delete_vhost_interface(self, idx):
        ret = self._vpp.api.delete_vhost_user_if(sw_if_index=idx)
        return ret.retval

    def get_vhostuser_interface(self, ifname):
        for iface in self.get_vhostuser_interfaces():
            if iface['ifname'] == ifname:
                return iface
        return None

    def get_vhostuser_interface_by_ifidx(self, idx):
        for iface in self.get_vhostuser_interfaces():
            if iface['sw_if_index'] == idx:
                return iface
        return None

    def get_vhostuser_interface_by_filename(self, sock_filename):
        for iface in self.get_vhostuser_interfaces():
            if iface['sock_filename'] == sock_filename:
                return iface
        return None

    def is_vhostuser_interface(self, ifname):
        for iface in self.get_vhostuser_interfaces():
            if iface['ifname'] == ifname:
                return True
        return False

    def is_vhostuser_interface_by_ifidx(self, idx):
        for iface in self.get_vhostuser_interfaces():
            if iface['sw_if_index'] == idx:
                return True
        return False

    # ##### RAWIP PLUGIN APIs #####

    def enable_rawip(self, idx, dst_mac):
        if not os.path.isfile('/usr/share/vpp/api/plugins/rawip.api.json'):
            return 2
        dst_hw_addr = mac_to_bytes(dst_mac)
        ret = self._vpp.api.rawip_enable_disable(
            sw_if_index=idx, dst_mac=dst_hw_addr, enable_disable=1)
        return ret.retval

    def disable_rawip(self, idx):
        if not os.path.isfile('/usr/share/vpp/api/plugins/rawip.api.json'):
            return 2
        ret = self._vpp.api.rawip_enable_disable(
            sw_if_index=idx, enable_disable=0)
        return ret.retval

    # ##### IPPASS PLUGIN APIs #####

    def enable_ippass(self, host_idx, wan_idx=0xffffffff, vm_idx=0xffffffff, rt_idx=0xffffffff):
        if not os.path.isfile('/usr/share/vpp/api/plugins/ippass.api.json'):
            return 2

        ret = self._vpp.api.ippass_enable_disable(host_if_index=host_idx,
                                                  wan_if_index=wan_idx,
                                                  vm_if_index=vm_idx,
                                                  rt_if_index=rt_idx,
                                                  enable_disable=1)
        return ret.retval

    def disable_ippass(self, host_idx, wan_idx=0xffffffff, vm_idx=0xffffffff, rt_idx=0xffffffff):
        if not os.path.isfile('/usr/share/vpp/api/plugins/ippass.api.json'):
            return 2

        ret = self._vpp.api.ippass_enable_disable(host_if_index=host_idx,
                                                  wan_if_index=wan_idx,
                                                  vm_if_index=vm_idx,
                                                  rt_if_index=rt_idx,
                                                  enable_disable=0)
        return ret.retval

    def ippass_check_status(self, host_idx):
        if not os.path.isfile('/usr/share/vpp/api/plugins/ippass.api.json'):
            return 2
        return self._vpp.api.ippass_if_status(host_if_index=host_idx)

    def add_del_ippass_permit_host_entry(self, host_idx, protocol, dir_out=1, dst_port=0xffffffff, ip=0, src_ip=0, add=1):
        if not os.path.isfile('/usr/share/vpp/api/plugins/ippass.api.json'):
            return 2

        proto = 0
        if "udp" == protocol:
            proto = 1
        elif "tcp" == protocol:
            proto = 2
        elif "icmp" == protocol:
            proto = 3
        elif "dhcp" == protocol:
            proto = 4
        elif "esp" == protocol:
            proto = 5
        else:
            print("Invalid protocol")
            return 1

        ipaddr = ipaddress.ip_address(ip)
        srcipaddr = ipaddress.ip_address(src_ip)
        ret = self._vpp.api.ippass_permit_host(host_if_index=host_idx,
                                               dst_port=dst_port,
                                               direction=dir_out,
                                               ip=ipaddr,
                                               src_ip=srcipaddr,
                                               protocol=proto,
                                               enable=add)
        return ret.retval

    def ippass_get_host_if_idx(self, vpp_if_id):
        if not os.path.isfile('/usr/share/vpp/api/plugins/ippass.api.json'):
            return 2

        host_if_id = None

        ret = self.ippass_check_status(vpp_if_id)
        if ret.retval == 0:
            return vpp_if_id

        bds = self.get_bridge_domains()
        for bd in bds:
            bd_ifs = list(map(lambda x: x['sw_if_index'], bd['sw_if_details']))
            if vpp_if_id in bd_ifs:
                bd_ifs.remove(vpp_if_id)
                if bd_ifs:
                    host_if_id = min(bd_ifs)
                    return host_if_id

        return host_if_id

    def enable_ippass_arp_termination(self, host_idx):
        if not os.path.isfile('/usr/share/vpp/api/plugins/ippass.api.json'):
            return 2

        ret = self._vpp.api.ippass_arp_term_enable_disable(host_if_index=host_idx,
                                                           enable_disable=1)
        return ret.retval

    def disable_ippass_arp_termination(self, host_idx):
        if not os.path.isfile('/usr/share/vpp/api/plugins/ippass.api.json'):
            return 2

        ret = self._vpp.api.ippass_arp_term_enable_disable(host_if_index=host_idx,
                                                           enable_disable=0)
        return ret.retval

    def enable_pppoe_ippass(self, host_idx, wan_idx=0xffffffff, vm_idx=0xffffffff):
        if not os.path.isfile('/usr/share/vpp/api/plugins/ippass.api.json'):
            return 2

        ret = self._vpp.api.ippass_pppoe_enable_disable(host_if_index=host_idx,
                                                        wan_if_index=wan_idx,
                                                        vm_if_index=vm_idx,
                                                        enable_disable=1)
        return ret.retval

    def disable_pppoe_ippass(self, host_idx, wan_idx=0xffffffff, vm_idx=0xffffffff):
        if not os.path.isfile('/usr/share/vpp/api/plugins/ippass.api.json'):
            return 2

        ret = self._vpp.api.ippass_pppoe_enable_disable(host_if_index=host_idx,
                                                        wan_if_index=wan_idx,
                                                        vm_if_index=vm_idx,
                                                        enable_disable=0)
        return ret.retval

    # ##### ROUTECTL PLUGIN APIs #####

    def enable_routectl(self, host_idx, rt_idx):
        if not os.path.isfile('/usr/share/vpp/api/plugins/routerctl.api.json'):
            return 1
        ret = self._vpp.api.rt_ctl_set(
            host_if=host_idx, rt_if=rt_idx, disable=0)
        return ret.retval

    def disable_routectl(self, host_idx, rt_idx):
        if not os.path.isfile('/usr/share/vpp/api/plugins/routerctl.api.json'):
            return 1
        ret = self._vpp.api.rt_ctl_set(
            host_if=host_idx, rt_if=rt_idx, disable=1)
        return ret.retval

    def routectl_set_ip(self, ip, disable=0):
        if not os.path.isfile('/usr/share/vpp/api/plugins/routerctl.api.json'):
            return 1
        try:
            ipaddr = ipaddress.ip_address(ip)
        except ValueError:
            ipaddr = ipaddress.ip_interface(ip).ip
        ret = self._vpp.api.rt_ctl_set_ip(ip=ipaddr, disable=disable)
        return ret.retval

    def routectl_set_mac(self, mac, disable=0):
        if not os.path.isfile('/usr/share/vpp/api/plugins/routerctl.api.json'):
            return 1
        macaddr = mac_to_bytes(mac)
        ret = self._vpp.api.rt_ctl_set_mac(mac=macaddr, disable=disable)
        return ret.retval

    def routectl_permit_proto(self, protocol, dst_addr, src_addr=0xffffffff, dst_port=0xffffffff, disable=0):
        if not os.path.isfile('/usr/share/vpp/api/plugins/routerctl.api.json'):
            return 1
        proto = 0
        if "udp" == protocol or 17 == protocol:
            proto = 17
        elif "gre" == protocol or 47 == protocol:
            proto = 47
        else:
            print("Invalid protocol")
            return 1
        dstip = ipaddress.ip_address(dst_addr)
        srcip = ipaddress.ip_address(src_addr)
        ret = self._vpp.api.rt_ctl_set_permit_proto(protocol=proto,
                                                    dst_addr=dstip,
                                                    src_addr=srcip,
                                                    dst_port=dst_port,
                                                    disable=disable)
        return ret.retval

    # ##### IP ROUTES APIs #####

    def add_del_vrf_table(self, table_id, is_add=True, is_ipv6=False):
        """brief Add / del table request
                A table can be added multiple times, but need be deleted only once.
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
        """
        if table_id == 0:
            return 1
        table = {'table_id': table_id, 'is_ip6': is_ipv6}
        ret = self._vpp.api.ip_table_add_del(table=table, is_add=is_add)
        return ret.retval

    def add_ip_route(self, ip_prefix, next_hop_address=None, is_multipath=0,
                     next_hop_sw_if_index=None, is_local=False, vrf=0, metric=0, check=True):
        """brief Add / del route request
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param is_multipath - Set to 1 if these paths will be added/removed
                        to/from the existing set, or 0 to replace
                        the existing set.
                        is_add=0 & is_multipath=0 implies delete all paths
                @param is_add - Are the paths being added or removed
        """

        is_ipv6 = False
        paths = []

        prefix = ipaddress.ip_interface(ip_prefix)
        is_ipv6 = True if prefix.ip.version == 6 else False
        if next_hop_address is not None:
            next_hop = ipaddress.ip_address(next_hop_address)
        else:
            next_hop = next_hop_address

        if check is True:
            if self.route_in_vrf(vrf, prefix, next_hop, next_hop_sw_if_index,
                                 is_ipv6, is_local):
                print('Route already exists.')
                return 0

        paths.append(self.encode_route_path(vrf, is_local, is_ipv6,
                                            next_hop_address=next_hop,
                                            next_hop_sw_if_index=next_hop_sw_if_index, metric=metric))
        route = self.encode_route(vrf, prefix, paths)
        ret = self._vpp.api.ip_route_add_del(
            is_add=1, is_multipath=is_multipath, route=route)
        return ret.retval

    def del_ip_route(self, ip_prefix, next_hop_address=None, is_multipath=0,
                     next_hop_sw_if_index=None, is_local=False, vrf=0, metric=0):

        is_ipv6 = False
        paths = []

        prefix = ipaddress.ip_interface(ip_prefix)
        is_ipv6 = True if prefix.ip.version == 6 else False
        if next_hop_address is not None:
            next_hop = ipaddress.ip_address(next_hop_address)
        else:
            next_hop = next_hop_address

        if False == self.route_in_vrf(vrf, prefix, next_hop, next_hop_sw_if_index,
                                      is_ipv6, is_local):
            print('Route does not exist.')
            return 0

        paths.append(self.encode_route_path(vrf, is_local, is_ipv6,
                                            next_hop_address=next_hop,
                                            next_hop_sw_if_index=next_hop_sw_if_index, metric=metric))
        route = self.encode_route(vrf, prefix, paths)
        ret = self._vpp.api.ip_route_add_del(
            is_add=0, is_multipath=is_multipath, route=route)
        return ret.retval

    def del_all_ip_routes_by_nh(self, next_hop_address=None, is_multipath=0,
                                next_hop_sw_if_index=None, is_local=False, vrf=0, is_ipv6=False):

        if next_hop_address == None and next_hop_sw_if_index == None:
            print('Next hop is not provided')
            return 1

        if next_hop_address is not None:
            next_hop = ipaddress.ip_address(next_hop_address)
        else:
            next_hop = next_hop_address

        table = {'table_id': vrf, 'is_ip6': is_ipv6}
        routes = self._vpp.api.ip_route_dump(table=table)
        for route in routes:
            rroute = route.route
            paths = rroute.paths

            if rroute.n_paths > 1:
                is_multipath = 1

            if ((next_hop or next_hop_sw_if_index) and rroute.table_id == vrf):
                (match, rpaths) = self.find_route_paths_by_nh(
                    paths, next_hop, next_hop_sw_if_index, is_ipv6)
                # print('Match: {}' .format(match))
                # print('Paths: {}' .format(rpaths))
                if match == True:
                    r = self.encode_route(
                        rroute.table_id, rroute.prefix, rpaths)
                    ret = self._vpp.api.ip_route_add_del(
                        is_add=0, is_multipath=is_multipath, route=r)

        return 0

    def del_ip_route_by_prefix(self, ip_prefix, is_multipath=0, is_local=False, vrf=0):
        if ip_prefix is None:
            return 1

        prefix = ip_prefix if (type(ip_prefix) == ipaddress.IPv4Interface or type(
            ip_prefix) == ipaddress.IPv6Interface) else ipaddress.ip_interface(ip_prefix)

        is_ipv6 = True if prefix.ip.version == 6 else False
        table = {'table_id': vrf, 'is_ip6': is_ipv6}
        routes = self._vpp.api.ip_route_dump(table=table)
        for route in routes:
            rpaths = []
            rroute = route.route
            paths = rroute.paths
            address = rroute.prefix.network_address
            address_len = rroute.prefix.prefixlen

            if rroute.n_paths > 1:
                is_multipath = 1

            if (address == prefix.ip and address_len == prefix._prefixlen):
                for p in paths:
                    rpaths.append(self.encode_route_path(vrf=vrf,
                                                         is_local=(
                                                             True if p.type == ROUTE_TYPE_LOCAL else False),
                                                         is_ipv6=is_ipv6,
                                                         next_hop_address=self.nexthop_ipaddr(
                                                             p, is_ipv6),
                                                         next_hop_sw_if_index=p.sw_if_index, metric=p.preference))

                    r = self.encode_route(
                        rroute.table_id, rroute.prefix, rpaths)
                    ret = self._vpp.api.ip_route_add_del(
                        is_add=0, is_multipath=is_multipath, route=r)

        return 0

    def encode_route_path_labels(self):
        """ Fill up/initialize the route's path label stack.
        """
        label_stack = {'is_uniform': 0,
                       'label': 0,
                       'ttl': 0,
                       'exp': 0}
        label_stack_list = []
        for i in range(16):
            label_stack_list.append(label_stack)

        return label_stack_list

    def encode_route_path(self, vrf, is_local, is_ipv6, next_hop_address=None,
                          next_hop_sw_if_index=None, metric=0):
        """Fill up/initialize the route's path.
                {'weight': 1,
                'preference': 0,
                'table_id': self.nh_table_id,
                'nh': self.nh.encode(),
                'next_hop_id': self.next_hop_id,
                'sw_if_index': self.nh_itf,
                'rpf_id': self.rpf_id,
                'proto': self.proto,
                'type': self.type,
                'flags': self.flags,
                'n_labels': len(self.nh_labels),
                'label_stack': self.encode_labels()}
        """
        label_stack_list = self.encode_route_path_labels()
        path = {'table_id': vrf,
                # MPLS labels 'n_labels': len(label_stack_list),
                'n_labels': 0,
                'label_stack': label_stack_list}

        # Type of route = local/normal
        if is_local:
            path['type'] = ROUTE_TYPE_LOCAL
        else:
            path['type'] = ROUTE_TYPE_NORMAL

        # IPv4/IPv6
        if is_ipv6:
            path['proto'] = PROTO_IPV6
        else:
            path['proto'] = PROTO_IPV4

        # Is there a next hop address?
        if next_hop_address:
            if not is_ipv6:
                path['nh'] = {'address': {'ip4': next_hop_address}}
            else:
                path['nh'] = {'address': {'ip6': next_hop_address}}

        # Is there a next hop intf index?
        if next_hop_sw_if_index:
            path['sw_if_index'] = next_hop_sw_if_index

        if metric != 0:
            path['preference'] = metric

        return path

    def encode_route(self, vrf, prefix, paths):
        """Fill up/initialize the route data structure.
        typedef ip_route
        {
                u32 table_id;
                u32 stats_index;
                vl_api_prefix_t prefix;
                u8 n_paths;
                vl_api_fib_path_t paths[n_paths];
        };
        """
        route = {'table_id': vrf,
                 'prefix': prefix if (type(prefix) == ipaddress.IPv4Network or type(prefix) == ipaddress.IPv6Network)
                 else prefix.network,
                 'n_paths': len(paths),
                 'paths': paths}
        return route

    def find_all_ip_routes_by_nh(self, vrf, next_hop_address, sw_if_index, is_lacal=False):
        """ Return all routes with next hop provided IP Address or if idx """
        rr = []
        is_ipv6 = True if next_hop_address.version == 6 else False

        table = {'table_id': vrf, 'is_ip6': is_ipv6}
        routes = self._vpp.api.ip_route_dump(table=table)
        for route in routes:
            rroutes = {}
            rpaths = []
            rroute = route.route
            paths = rroute.paths

            if (next_hop_address and rroute.table_id == vrf
                    and (next_hop_address in [self.nexthop_ipaddr(p, is_ipv6) for p in paths])):
                (match, rpaths) = self.find_route_paths_by_nh(
                    paths, next_hop_address, sw_if_index, is_ipv6)
                if match == True:
                    rr.append(rroute)
                    rroutes['route'] = rroute
                    rroutes['paths'] = rpaths
                    rr.append(rroutes)

            elif (sw_if_index and route.table_id == vrf
                  and sw_if_index in [p.sw_if_index for p in paths]):
                (match, rpaths) = self.find_route_paths_by_nh(
                    paths, next_hop_address, sw_if_index, is_ipv6)
                if match == True:
                    rroutes['route'] = rroute
                    rroutes['paths'] = rpaths
                    rr.append(rroutes)
        return rr

    def route_in_vrf(self, vrf, ip_prefix,
                     next_hop_address, sw_if_index, is_ipv6=False,
                     is_local=False):
        """Returns True, if the route if present in the VRF.

        Pulls the VPP FIB to see if the route is present in the VRF.
        The route is identified by the tuple,
        (ip_address, prefixlen, next_hop_address)
        If the route is present, returns True or else returns False.
        """

        table = {'table_id': vrf, 'is_ip6': is_ipv6}
        routes = self._vpp.api.ip_route_dump(table=table)

        ip = ip_prefix.ip
        prefixlen = ip_prefix._prefixlen

        for route in routes:
            route = route.route
            paths = route.paths
            table_id = route.table_id
            address = route.prefix.network_address
            address_len = route.prefix.prefixlen

            # if there's a valid next_hop_address check for the route by
            # including it
            if (next_hop_address and table_id == vrf and
                    address_len == prefixlen and
                    address == ip and
                    (next_hop_address in [self.nexthop_ipaddr(p, is_ipv6) for p in paths])):
                print('Route: {}/{} to {} exists in VRF:{}'
                      .format(ip, prefixlen, next_hop_address, vrf))
                return True
            elif (sw_if_index and table_id == vrf and
                    address_len == prefixlen and
                    address == ip and
                    sw_if_index in [p.sw_if_index for p in paths]):
                print('Route: {}/{} to sw_if_idx:{} is imported into VRF:{}'
                      .format(ip, prefixlen, sw_if_index, vrf))
                return True
            elif (is_local and table_id == vrf and
                    address_len == prefixlen and
                    address == ip and self.any_local_routes(paths)):
                print('Local route: %s/%s exists in VRF:%s',
                      ip, prefixlen, vrf)
                return True
        else:
            print('Route: {}/{} to {} does not exist in VRF:{}'
                  .format(ip, prefixlen, next_hop_address, vrf))
            return False

    def nexthop_ipaddr(self, p, is_ipv6):
        # Return the ipaddress.IPvXAddress in the route path 'p'
        addr = p.nh.address.ip6 if is_ipv6 else p.nh.address.ip4
        return addr

    def any_local_routes(self, paths):
        return any((p.type == ROUTE_TYPE_LOCAL for p in paths))

    def find_route_paths_by_nh(self, route_paths, next_hop, sw_if_index, is_ipv6=False):
        paths = []
        match = False

        for p in route_paths:
            if ((next_hop != None and next_hop == self.nexthop_ipaddr(p, is_ipv6))
                    or (sw_if_index != None and sw_if_index == p.sw_if_index)):
                match = True
                paths.append(self.encode_route_path(vrf=p.table_id,
                                                    is_local=(
                                                        True if p.type == ROUTE_TYPE_LOCAL else False),
                                                    is_ipv6=is_ipv6,
                                                    next_hop_address=self.nexthop_ipaddr(
                                                        p, is_ipv6),
                                                    next_hop_sw_if_index=p.sw_if_index, metric=p.preference))

        return (match, paths)

    def add_ip_neighbor(self, sw_if_index, mac, ip):
        macaddr = mac_to_bytes(mac)
        ipaddr = ipaddress.ip_address(ip)
        neighbor = {'sw_if_index': sw_if_index,
                    'mac_address': macaddr,
                    'ip_address': ipaddr}

        ret = self._vpp.api.ip_neighbor_add_del(is_add=1, neighbor=neighbor)
        return ret.retval

    def del_ip_neighbor(self, sw_if_index, mac, ip):
        macaddr = mac_to_bytes(mac)
        ipaddr = ipaddress.ip_address(ip)
        neighbor = {'sw_if_index': sw_if_index,
                    'mac_address': macaddr,
                    'ip_address': ipaddr}

        ret = self._vpp.api.ip_neighbor_add_del(is_add=0, neighbor=neighbor)
        return ret.retval

    def get_ip_neighbors(self):
        vaf = VppEnum.vl_api_address_family_t
        ret = self._vpp.api.ip_neighbor_dump(sw_if_index=0xFFFFFFFF, af=vaf.ADDRESS_IP4)

        for neighbor_detail in ret:
            neighbor = neighbor_detail.neighbor
            sw_if_index = neighbor.sw_if_index
            mac_address = neighbor.mac_address
            ip_address = neighbor.ip_address
            yield sw_if_index, mac_address, ip_address

    def get_ip_neighbors_by_if_index(self, sw_if_index):
        neighbors = []
        for sw_idx, mac_address, ip_address in self.get_ip_neighbors():
            if sw_idx == sw_if_index:
                neighbors_info = {'sw_if_index': sw_idx, 'mac_address': mac_address, 'ip_address': ip_address}
                neighbors.append(neighbors_info)
        return neighbors

    def add_l2fib_entry(self, mac, bridge_idx, sw_if_index):
        """Add L2 FIB entry request
                    @param client_index - opaque cookie to identify the sender
                    @param context - sender context, to match reply w/ request
                    @param mac - the entry's mac address
                    @param bd_id - the entry's bridge domain id
                    @param sw_if_index - the interface
                    @param is_add - If non zero add the entry, else delete it
                    @param static_mac -
                    @param filter_mac -
                    @param bvi_mac -
        """
        macaddr = mac_to_bytes(mac)
        ret = self._vpp.api.l2fib_add_del(mac=macaddr,
                                        bd_id=bridge_idx, 
                                        sw_if_index=sw_if_index,
                                        is_add=1)
        return ret.retval

    # ##### VXLAN TUNNELS #####
    def vxlan_create(self, vni, lvtep, rvtep, instance=0xffffffff, mcast_if_index=0xffffffff):
        """Create or delete a VXLAN tunnel
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param is_add - Use 1 to create the tunnel, 0 to remove it
                @param instance - optional unique custom device instance, else ~0.
                @param src_address - Source IP address
                @param dst_address - Destination IP address, can be multicast
                @param mcast_sw_if_index - Interface for multicast destination
                @param encap_vrf_id - Encap route table FIB index
                @param decap_next_index - index of decap next graph node
                @param vni - The VXLAN Network Identifier, uint24
        """
        local_ip = ipaddress.ip_address(lvtep)
        remote_ip = ipaddress.ip_address(rvtep)
        ret = self._vpp.api.vxlan_add_del_tunnel(vni=vni,
                                                 src_address=local_ip,
                                                 dst_address=remote_ip,
                                                 instance=instance,
                                                 decap_next_index=0xffffffff,
                                                 mcast_sw_if_index=mcast_if_index,
                                                 #  is_ipv6=1 if local_ip.version == 6 else False,
                                                 is_add=1)
        return (ret.retval, ret.sw_if_index)

    def vxlan_delete(self, vni, lvtep, rvtep, instance=0xffffffff, mcast_if_index=0xffffffff):
        """Create or delete a VXLAN tunnel
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param is_add - Use 1 to create the tunnel, 0 to remove it
                @param instance - optional unique custom device instance, else ~0.
                @param src_address - Source IP address
                @param dst_address - Destination IP address, can be multicast
                @param mcast_sw_if_index - Interface for multicast destination
                @param encap_vrf_id - Encap route table FIB index
                @param decap_next_index - index of decap next graph node
                @param vni - The VXLAN Network Identifier, uint24
        """
        local_ip = ipaddress.ip_address(lvtep)
        remote_ip = ipaddress.ip_address(rvtep)
        ret = self._vpp.api.vxlan_add_del_tunnel(vni=vni,
                                                 src_address=local_ip,
                                                 dst_address=remote_ip,
                                                 instance=instance,
                                                 decap_next_index=0xffffffff,
                                                 mcast_sw_if_index=mcast_if_index,
                                                 #  is_ipv6=1 if local_ip.version == 6 else False,
                                                 is_add=0)
        return ret.retval

    def vxlan_delete_by_idx(self, idx):
        iface = self.get_vxlan_if_by_ifidx(idx)
        if not iface or iface == None:
            return 1
        ret = self.vxlan_delete(vni=iface.get('vni'),
                                lvtep=ipaddress.ip_address(
                                    iface.get('local_address')),
                                rvtep=ipaddress.ip_address(
                                    iface.get('remote_address')),
                                instance=iface.get('instance', 0xffffffff),
                                mcast_if_index=iface.get('mcast_sw_if_index', 0xffffffff))
        return ret

    def get_vxlan_interfaces(self):
        ret = self._vpp.api.vxlan_tunnel_dump(sw_if_index=0xffffffff)
        for iface in ret:
            yield {'sw_if_index': iface.sw_if_index,
                   'vni': iface.vni,
                   'instance': iface.instance,
                   'local_address': iface.src_address,
                   'remote_address': iface.dst_address,
                   # 'local_address': bytes_to_ip(iface.src_address, iface.is_ipv6),
                   # 'remote_address': bytes_to_ip(iface.dst_address, iface.is_ipv6),
                   # 'is_ipv6': iface.is_ipv6,
                   'mcast_sw_if_index': iface.mcast_sw_if_index,
                   'encap_vrf_id': iface.encap_vrf_id
                   }

    def get_vxlan_if_by_ifidx(self, idx):
        for iface in self.get_vxlan_interfaces():
            if iface['sw_if_index'] == idx:
                return iface
        return None

    # ##### VXLAN GPE TUNNELS #####

    def vxlan_gpe_create(self, vni, lvtep, rvtep, mcast_if_index=0xffffffff):
        """define vxlan_gpe_add_del_tunnel
                {
                u32 client_index;
                u32 context;
                vl_api_address_t local;
                vl_api_address_t remote;
                vl_api_interface_index_t mcast_sw_if_index;
                u32 encap_vrf_id;
                u32 decap_vrf_id;
                vl_api_ip_proto_t protocol;
                u32 vni;
                bool is_add [default=true];
                };
        """
        local_ip = ipaddress.ip_address(lvtep)
        remote_ip = ipaddress.ip_address(rvtep)
        ret = self._vpp.api.vxlan_gpe_add_del_tunnel(vni=vni,
                                                     local=local_ip,
                                                     remote=remote_ip,
                                                     #  instance=instance,
                                                     mcast_sw_if_index=mcast_if_index,
                                                     #  is_ipv6=1 if local_ip.version == 6 else False,
                                                     is_add=1)
        return (ret.retval, ret.sw_if_index)

    def vxlan_gpe_delete(self, vni, lvtep, rvtep, mcast_if_index=0xffffffff):
        """define vxlan_gpe_add_del_tunnel
                {
                u32 client_index;
                u32 context;
                vl_api_address_t local;
                vl_api_address_t remote;
                vl_api_interface_index_t mcast_sw_if_index;
                u32 encap_vrf_id;
                u32 decap_vrf_id;
                vl_api_ip_proto_t protocol;
                u32 vni;
                bool is_add [default=true];
                };
        """
        local_ip = ipaddress.ip_address(lvtep)
        remote_ip = ipaddress.ip_address(rvtep)
        ret = self._vpp.api.vxlan_gpe_add_del_tunnel(vni=vni,
                                                     local=local_ip,
                                                     remote=remote_ip,
                                                     #  instance=instance,
                                                     mcast_sw_if_index=mcast_if_index,
                                                     #  is_ipv6=1 if local_ip.version == 6 else False,
                                                     is_add=0)
        return ret.retval

    def vxlan_gpe_delete_by_idx(self, idx):
        iface = self.get_vxlan_gpe_if_by_ifidx(idx)
        if not iface or iface == None:
            return 1
        ret = self.vxlan_gpe_delete(vni=iface.get('vni'),
                                    lvtep=ipaddress.ip_address(
                                        iface.get('local_address')),
                                    rvtep=ipaddress.ip_address(
                                        iface.get('remote_address')),
                                    mcast_if_index=iface.get('mcast_sw_if_index', 0xffffffff))
        return ret

    def get_vxlan_gpe_interfaces(self):
        ret = self._vpp.api.vxlan_gpe_tunnel_dump(sw_if_index=0xffffffff)
        for iface in ret:
            yield {'sw_if_index': iface.sw_if_index,
                   'vni': iface.vni,
                   'local_address': iface.local,
                   'remote_address': iface.remote,
                   # 'is_ipv6': iface.is_ipv6,
                   'mcast_sw_if_index': iface.mcast_sw_if_index,
                   'encap_vrf_id': iface.encap_vrf_id
                   }

    def get_vxlan_gpe_if_by_ifidx(self, idx):
        for iface in self.get_vxlan_gpe_interfaces():
            if iface['sw_if_index'] == idx:
                return iface
        return None

    def ip6_ra_config(self, sw_if_index, no=0, suppress=0):
        """Configure IPv6 RA suppress on the VPP interface."""
        """define sw_interface_ip6nd_ra_config
			{
       		...
			u8 no;  -  negate state
			u8 suppress; - suppress RA sending;
			};
		"""
        ret = self._vpp.api.sw_interface_ip6nd_ra_config(
            sw_if_index=sw_if_index,
            is_no=no,
            suppress=suppress)
        return ret.retval

    # ##### DRCTL APIs #####
    def enable_drctl(self, host_idx, mac):
        if not os.path.isfile(VPP_JSON_DIR+'plugins/drctl.api.json'):
            print("DI not file {}" .format(VPP_JSON_DIR+'drctl.api.json'))
            return 1

        ret = self._vpp.api.drctl_set(
            host_if=host_idx, mac=mac_to_bytes(mac), disable=0)
        return ret.retval

    def disable_drctl(self, host_idx, mac):
        if not os.path.isfile(VPP_JSON_DIR+'plugins/drctl.api.json'):
            return 1

        ret = self._vpp.api.drctl_set(
            host_if=host_idx, mac=mac_to_bytes(mac), disable=1)
        return ret.retval

    def drctl_set_ip(self, host_idx, ip, delete=0):
        if not os.path.isfile(VPP_JSON_DIR+'plugins/drctl.api.json'):
            return 1
        try:
            ipaddr = ipaddress.ip_address(ip)
        except ValueError:
            ipaddr = ipaddress.ip_interface(ip).ip
        ret = self._vpp.api.drctl_set_ip(
            host_if=host_idx, ip=ipaddr, delete=delete)
        return ret.retval

    def drctl_set_arp(self, host_idx, rtif_idx, delete=0):
        if not os.path.isfile(VPP_JSON_DIR+'plugins/drctl.api.json'):
            return 1

        ret = self._vpp.api.drctl_set_arp(
            dupl_if=host_idx, rt_if=rtif_idx, delete=delete)
        return ret.retval

    def drctl_set_dhcp(self, host_idx, rtif_idx, delete=0):
        if not os.path.isfile(VPP_JSON_DIR+'plugins/drctl.api.json'):
            return 1

        ret = self._vpp.api.drctl_set_dhcp(
            host_if=host_idx, rt_if=rtif_idx, delete=delete)
        return ret.retval

    # ##### L2 EMULATION APIs #####
    def l2_emulation_enable_disable(self, idx, enable=1):
        """ brief L2 emulation at L3
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - interface the operation is applied to
                @param enable - Turn the service on or off
        """
        if not os.path.isfile(VPP_JSON_DIR+'plugins/l2e.api.json'):
            return 1

        ret = self._vpp.api.l2_emulation(sw_if_index=idx, enable=enable)
        return ret.retval

    # ##### PPPoE CLIENT APIs #####
    def pppoeclient_add_client(self, eth_ifidx, host_uniq=1111, ac_name=0, idle_timeout=0, retry_timeout=0):
        """ brief Set or delete an PPPOE client
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param is_add - add address if non-zero, else delete
                @param sw_if_index - client bound ethernet if index
                @param host_uniq - tag identifying a PPPOE client on the the sw_if_index
                @param ac_name - prefered AC to connect to
        """
        if not os.path.isfile(VPP_JSON_DIR+'plugins/pppoeclient.api.json'):
            return 1

        ret = self._vpp.api.pppoe_add_del_client(sw_if_index=eth_ifidx,
                                                 host_uniq=host_uniq,
                                                 ac_name=0 if ac_name == 0 else binary_type(
                                                     ac_name),
                                                 idle_timeout=idle_timeout,
                                                 retry_timeout=retry_timeout,
                                                 is_add=1)
        return (ret.retval, ret.pppox_sw_if_index)

    def pppoeclient_del_client(self, eth_ifidx, host_uniq=1111):
        """ brief Set or delete an PPPOE client
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param is_add - add address if non-zero, else delete
                @param sw_if_index - client bound ethernet if index
                @param host_uniq - tag identifying a PPPOE client on the the sw_if_index
        """
        if not os.path.isfile(VPP_JSON_DIR+'plugins/pppoeclient.api.json'):
            return 1

        ret = self._vpp.api.pppoe_add_del_client(
            sw_if_index=eth_ifidx, host_uniq=host_uniq, is_add=0)
        return ret.retval

    def pppoeclient_set_auth(self, ppp_ifidx, username, password, auth_type='pap'):
        """ Set pppox auth used by ppp (can be used by pap/chap and etc.)
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - pppox software if index
                @param username - ppp username
                @param password - ppp password
        @param auth_type - PAP / CHAP authentication type
        """
        if not os.path.isfile(VPP_JSON_DIR+'plugins/pppox.api.json'):
            return 1

        atype = 0
        if "pap" == auth_type or "PAP" == auth_type or auth_type == 1:
            atype = 1
        elif "chap" == auth_type or "CHAP" == auth_type or auth_type == 2:
            atype = 2
        ret = self._vpp.api.pppox_set_auth(sw_if_index=ppp_ifidx,
                                           username=binary_type(username),
                                           password=binary_type(password),
                                           auth_type=atype)
        return ret.retval

    def pppoeclient_set_host(self, ppp_ifidx, host_idx=0xffffffff, metric=0xff, is_add=True):
        if not os.path.isfile(VPP_JSON_DIR+'plugins/pppox.api.json'):
            return 1

        ret = self._vpp.api.pppox_set_host(pppox_if=ppp_ifidx,
                                           host_if=host_idx,
                                           metric=metric,
                                           delete=1 if is_add == False else 0)
        return ret.retval

    def get_pppoe_client_ifs(self):
        ret = self._vpp.api.pppoe_client_dump(sw_if_index=0xffffffff)
        for iface in ret:
            ac_mac = bytearray(iface.ac_mac)
            yield {'pppox_sw_if_index': iface.pppox_sw_if_index,
                   'ethif_sw_if_index': iface.ethif_sw_if_index,
                   'host_uniq': iface.host_uniq,
                   'session_id': iface.session_id,
                   'state': pppoe_states.get(iface.state),
                   'ac_name': fix_string(iface.ac_name),
                   'ac_mac_addr': ':'.join(["%02x" % int(c) for c in ac_mac]),
                   'err_code': iface.err_code,
                   'err_reason': fix_string(iface.err_reason)
                   }

    def get_pppoe_client_by_pppox_ifidx(self, idx):
        for iface in self.get_pppoe_client_ifs():
            if iface['pppox_sw_if_index'] == idx:
                return iface
        return None

    def get_pppox_conns(self):
        ret = self._vpp.api.pppox_session_dump(sw_if_index=0xffffffff)
        for iface in ret:
            yield {'pppox_sw_if_index': iface.pppox_sw_if_index,
                   'host_if_index': iface.host_if_index,
                   'local_address': bytes_to_ip(iface.local_address, iface.is_ipv6),
                   'peer_address': bytes_to_ip(iface.peer_address, iface.is_ipv6),
                   'is_ipv6': iface.is_ipv6,
                   'err_code': iface.err_code,
                   'err_reason': fix_string(iface.err_reason)
                   }

    def get_pppox_conn_by_pppox_ifidx(self, idx):
        for iface in self.get_pppox_conns():
            if iface['pppox_sw_if_index'] == idx:
                return iface
        return None

    def get_pppox_conn_by_host_ifidx(self, idx):
        for iface in self.get_pppox_conns():
            if iface['host_if_index'] == idx:
                return iface
        return None

    def get_pppox_idx_by_host_ifidx(self, idx):
        for iface in self.get_pppox_conns():
            if iface['host_if_index'] == idx:
                return iface['pppox_sw_if_index']
        return None

    # ##### Port Mirroring APIs #####
    def enable_port_mirroring(self, src_idx, dst_idx, direction=3, is_l2=0):
        """	Enable/Disable span to mirror traffic from one interface to another
                @param client_index - opaque cookie to identify the sender
                @param context - sender context which was passed in the request
                @param sw_if_index_from - interface to be mirrored
                @param sw_if_index_to - interface where the traffic is mirrored
                @param state - 0 = disabled, 1 = rx enabled, 2 = tx enabled, 3 tx & rx enabled
                @param is_l2 - 0 = mirror at hw device level, 1 = mirror at L2
"""
        ret = self._vpp.api.sw_interface_span_enable_disable(sw_if_index_from=src_idx,
                                                             sw_if_index_to=dst_idx,
                                                             is_l2=is_l2,
                                                             state=direction)
        return ret.retval

    def disable_port_mirroring(self, src_idx, dst_idx, is_l2=0):
        ret = self._vpp.api.sw_interface_span_enable_disable(sw_if_index_from=src_idx,
                                                             sw_if_index_to=dst_idx,
                                                             is_l2=is_l2,
                                                             state=0)
        return ret.retval

    # ##### WireGuard APIs #####
    def wg_generate_keys(self):
        ret = self._vpp.api.wireguard_generate_keys_base64()
        return (ret.retval, fix_string(ret.private_key), fix_string(ret.public_key))

    def wg_create_interface(self, src_ip, port, instance=None, private_key=None, public_key=None, private_key_base64=None, public_key_base64=None, 
                            generate_key=False, base64=True, cfg_instance=0):
        """ Create wireguard interface
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param private_key - private key in binary format of this device
                @param public_key - public key in binary format of this device
                @param port - port of this device
                @param src_ip - packet sent through this interface us this
                                address as the IP source.

            define wireguard_interface_create {
                u32 client_index;
                u32 context;
                vl_api_wireguard_interface_t interface;
                bool generate_key;
            };                
            
            typedef wireguard_interface
            {
                u32 user_instance [default=0xffffffff];
                vl_api_interface_index_t sw_if_index;
                u8 private_key[32];
                u8 public_key[32];
                u8 private_key_base64[45];
                u8 public_key_base64[45];
                u16 port;
                vl_api_address_t src_ip;
            };
        """
        if not src_ip or not port:
            return 1, 0xffffffff
        if base64 and not private_key_base64:
            return 1, 0xffffffff
        wg_interface = {
                        #'user_instance': int(instance),
                        'sw_if_index': 0,
                        'private_key': private_key,
                        'public_key': public_key,
                        'private_key_base64': binary_type(private_key_base64),
                        'public_key_base64': None if not public_key_base64 else binary_type(public_key_base64),
                        'port': int(port),
                        'src_ip': src_ip,
                        'config_instance': cfg_instance}
        
        ret = self._vpp.api.wireguard_interface_create(interface=wg_interface, generate_key=generate_key, base64=base64)
        
        return (ret.retval, ret.sw_if_index)

    def wg_delete_interface(self, sw_if_index):
        """ autoreply define wireguard_interface_delete
            {
            u32 client_index;
            u32 context;
            vl_api_interface_index_t sw_if_index;
            };
        """
        ret = self._vpp.api.wireguard_interface_delete(sw_if_index=sw_if_index)
        return ret.retval

    def wg_get_interfaces(self,show_private_key=True):
        """ define wireguard_interface_dump
            {
            u32 client_index;
            u32 context;
            bool show_private_key;
            vl_api_interface_index_t sw_if_index;
            };
            typedef wireguard_interface
            {
                u32 user_instance [default=0xffffffff];
                vl_api_interface_index_t sw_if_index;
                u8 private_key[32];
                u8 public_key[32];
                u8 private_key_base64[45];
                u8 public_key_base64[45];
                u16 port;
                vl_api_address_t src_ip;
            };
        """
        ret = self._vpp.api.wireguard_interface_dump(show_private_key=show_private_key)
        for iface in ret:
            yield {'instance': iface.interface.user_instance,
                   'ifname': 'wg{}' .format(iface.interface.user_instance),
                   'sw_if_index': iface.interface.sw_if_index,
                   'local_endpoint': iface.interface.src_ip,
                   'local_port': iface.interface.port,
                   'private_key': fix_string(iface.interface.private_key_base64),
                   'public_key': fix_string(iface.interface.public_key_base64),
                   'cfg_idx': iface.interface.config_instance }

    def wg_get_interface_by_ifname(self, ifname):
        for iface in self.wg_get_interfaces():
            if iface['ifname'] == ifname:
                return iface
        return None

    def wg_get_interface_by_local_port(self, local_port):
        for iface in self.wg_get_interfaces():
            if iface['local_port'] == local_port:
                return iface
        return None

    def wg_get_interface_by_ifidx(self, idx):
        for iface in self.wg_get_interfaces():
            if iface['sw_if_index'] == idx:
                return iface
        return None

    def wg_get_interface_by_cfg_instance(self, instance):
        for iface in self.wg_get_interfaces():
            if iface['cfg_idx'] == instance:
                return iface['sw_if_index']
        return None

    def wg_get_interface_ifidx_by_cfg_instance(self, instance):
        for iface in self.wg_get_interfaces():
            if iface['cfg_idx'] == instance:
                return iface['sw_if_index']
        return None

    def is_wireguard_interface(self, ifname):
        for iface in self.wg_get_interfaces():
            if iface['ifname'] == ifname:
                return True
        return False

    def is_wg_interface_by_ifidx(self, idx):
        for iface in self.wg_get_interfaces():
            if iface['sw_if_index'] == idx:
                return True
        return False

    def wg_add_peer(self, wgif, endpoint, port, allowed_ips, public_key=None, public_key_base64=None, base64=True, keepalive=0):
        """ define wireguard_peer_add
            {
                u32 client_index;
                u32 context;
                bool base64;
                vl_api_wireguard_peer_t peer; 
            };
            
            define wireguard_peer_add_reply
            {
                u32 context;
                i32 retval;
                u32 peer_index;
            };
            
            typedef wireguard_peer
            {
                u8 public_key[32];
                u8 public_key_base64[45];
                u16 port;
                u16 persistent_keepalive;
                u32 table_id;
                vl_api_address_t endpoint;
                vl_api_interface_index_t sw_if_index;
                vl_api_wireguard_peer_flags_t flags;
                u8 n_allowed_ips;
                vl_api_prefix_t allowed_ips[n_allowed_ips];
            };
        """
        if not wgif or not endpoint or not port:
            return 1, 0xffffffff
        ips = []
        for i in allowed_ips:
            ii = ipaddress.ip_interface(i)
            ips.append(ii.network)

        peer = {'sw_if_index': wgif,
                'public_key': public_key,
                'public_key_base64': None if not public_key_base64 else binary_type(public_key_base64),
                'port': int(port),
                'endpoint': endpoint,
                'persistent_keepalive': keepalive,
                'allowed_ips': ips,
                'n_allowed_ips': len(ips)
                }
        
        ret = self._vpp.api.wireguard_peer_add(peer=peer, base64=base64)
        return (ret.retval, ret.peer_index)

    def wg_del_peer(self, peer_idx):
        """ brief Remove peer 
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param peer_index
        """
        ret = self._vpp.api.wireguard_peer_remove(peer_index=peer_idx)
        return ret.retval

    def wg_get_peers(self,peer_index=0xffffffff):
        """ brief Dump all peers
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param peer_index - peer index to be dumped.  If 0xFFFFFFFF dumps all peers
            
            typedef wireguard_peer
            {
                u8 public_key[32];
                u8 public_key_base64[45];
                u16 port;
                u16 persistent_keepalive;
                u32 table_id;
                vl_api_address_t endpoint;
                vl_api_interface_index_t sw_if_index;
                vl_api_wireguard_peer_flags_t flags;
                u8 n_allowed_ips;
                vl_api_prefix_t allowed_ips[n_allowed_ips];
            };
        """
        ret = self._vpp.api.wireguard_peers_dump(peer_index=peer_index)
        for iface in ret:
            yield {'wgif': iface.peer.sw_if_index,
                   'peer_endpoint': iface.peer.endpoint,
                   'peer_port': iface.peer.port,
                   'persistent_keepalive': iface.peer.persistent_keepalive,
                   'table_id': iface.peer.table_id,
                   'public_key': fix_string(iface.peer.public_key_base64),
                   'n_allowed_ips': iface.peer.n_allowed_ips,
                   'allowed_ips': iface.peer.allowed_ips}
    
    def wg_get_peer_by_endpoint(self, endpoint):
        for iface in self.wg_get_peers():
            if str(iface['peer_endpoint']) == endpoint:
                return iface
        return None

    # ##### VRRP APIs #####

    def vrrp_vr_add_or_del(self, sw_if_index, vr_id, is_add, priority, vip_ip, is_ipv6=0, interval=100):
        if not os.path.isfile('/usr/share/vpp/api/plugins/vrrp.api.json'):
            return 2
        """
            enum vrrp_vr_flags
            {
              VRRP_API_VR_PREEMPT = 0x1,
              VRRP_API_VR_ACCEPT = 0x2,
              VRRP_API_VR_UNICAST = 0x4,
              VRRP_API_VR_IPV6 = 0x8,
            };
            /** \brief VRRP: Add or delete a VRRP virtual router
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param is_add - 0 if deleting, != 0 if adding
                @param sw_if_index - interface backed up by this vr
                @param vr_id - the VR ID advertised by this vr
                @param priority - the priority advertised for this vr
                @param interval - interval between advertisements in centiseconds
                @param flags - bit flags for booleans - preempt, accept, unicast, ipv6;
                @param n_addrs - number of addresses being backed up by this vr
                @param addrs - the addresses (VIPs) backed up by this vr
            */
            autoreply define vrrp_vr_add_del {
              u32 client_index;
              u32 context;
              u8 is_add;
              vl_api_interface_index_t sw_if_index;
              u8 vr_id;
              u8 priority;
              u16 interval;
              vl_api_vrrp_vr_flags_t flags;
              u8 n_addrs;
              vl_api_address_t addrs[n_addrs];
            };
        """

        priority = int(priority)
        vip_list = vip_ip.split(",")
        n_addrs = len(vip_list)

        if is_ipv6 != 0:
            flags = VppEnum.vl_api_vrrp_vr_flags_t.VRRP_API_VR_IPV6 | VppEnum.vl_api_vrrp_vr_flags_t.VRRP_API_VR_UNICAST | VppEnum.vl_api_vrrp_vr_flags_t.VRRP_API_VR_PREEMPT
        else:
            flags = VppEnum.vl_api_vrrp_vr_flags_t.VRRP_API_VR_UNICAST | VppEnum.vl_api_vrrp_vr_flags_t.VRRP_API_VR_PREEMPT

        ret = self._vpp.api.vrrp_vr_add_del(is_add=is_add,
                                            sw_if_index=sw_if_index,
                                            vr_id=vr_id,
                                            priority=priority,
                                            interval=interval,
                                            flags=flags,
                                            n_addrs=n_addrs,
                                            addrs=vip_list)
        return ret.retval

    def vrrp_vr_delete(self, vr_id):
        if not os.path.isfile('/usr/share/vpp/api/plugins/vrrp.api.json'):
            return 2
        """
            /**
             * @brief Delete an existing VRRP instance
             * @param client_index - opaque cookie to identify the sender
             * @param context - returned sender context, to match reply w/ request
             * @param vrrp_index - index of the VRRP instance to delete
             */
            autoreply define vrrp_vr_del {
              u32 client_index;
              u32 context;
              u32 vrrp_index;
            };
        """

        ret = self._vpp.api.vrrp_vr_del(vrrp_index=vr_id)
        return ret.retval

    def vrrp_vr_start_stop(self, start_stop, sw_if_index, vr_id, is_ipv6=0):
        if not os.path.isfile('/usr/share/vpp/api/plugins/vrrp.api.json'):
            return 2
        """
            /** \brief VRRP: start or shutdown the VRRP protocol for a virtual router
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - interface ID that VR is backing up
                @param vr_id - VR ID
                @param is_ipv6 - 1 for IPv6, 0 for IPv4
                @param is_start - 1 to start VRRP proto on this VR, 0 to shutdown
            */
            autoreply define vrrp_vr_start_stop {
              u32 client_index;
              u32 context;
              vl_api_interface_index_t sw_if_index;
              u8 vr_id;
              u8 is_ipv6;
              u8 is_start;
            };
        """

        is_start = 1 if start_stop == 'start' else 0

        ret = self._vpp.api.vrrp_vr_start_stop(sw_if_index=sw_if_index,
                                               vr_id=vr_id,
                                               is_ipv6=is_ipv6,
                                               is_start=is_start)
        return ret.retval

    def vrrp_vr_peers_set(self, sw_if_index, vr_id, peer_ip, is_ipv6=0):
        if not os.path.isfile('/usr/share/vpp/api/plugins/vrrp.api.json'):
            return 2

        """
            /** \brief VRRP: set unicast peers for a VR
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - interface ID that VR is backing up
                @param vr_id - VR ID
                @param is_ipv6 - 1 for IPv6, 0 for IPv4
                @param n_addrs - number of peer addresses
                @param addrs - peer addresses
            */
            autoreply define vrrp_vr_set_peers {
              u32 client_index;
              u32 context;
              vl_api_interface_index_t sw_if_index;
              u8 vr_id;
              u8 is_ipv6;
              u8 n_addrs;
              vl_api_address_t addrs[n_addrs];
            };
        """

        peer_list = peer_ip.split(",")
        n_addrs = len(peer_list)

        ret = self._vpp.api.vrrp_vr_set_peers(sw_if_index=sw_if_index,
                                              vr_id=vr_id,
                                              is_ipv6=is_ipv6,
                                              n_addrs=n_addrs,
                                              addrs=peer_list)
        return ret.retval

    def vrrp_vr_track_if_add_del(self, sw_if_index, vr_id, tr_intfs, is_add=1, is_ipv6=0):
        if not os.path.isfile('/usr/share/vpp/api/plugins/vrrp.api.json'):
            return 2
        """
            /** \brief VR interface tracking
                @param sw_if_index - the interface index to track (not the VR sw_if_index)
                @param priority - the adjustment to VR priority if intf is down
            */
            typedef vrrp_vr_track_if
            {
              vl_api_interface_index_t sw_if_index;
              u8 priority;
            };

            /** \brief VRRP: Add/delete VR priority tracking of interface status
                @param context - sender context which was passed in the request
                @param sw_if_index - interface index
                @param is_ipv6 - 0 -> IPv4, 1 -> IPv6
                @param vr_id - ID of VR
                @param is_add - 0 -> delete, 1 -> add
                @param n_ifs - number of interface tracking records
                @param ifs - array of interface tracking records
            */
            autoreply define vrrp_vr_track_if_add_del
            {
              u32 client_index;
              u32 context;
              vl_api_interface_index_t sw_if_index;
              u8 is_ipv6;
              u8 vr_id;
              u8 is_add;
              u8 n_ifs;
              vl_api_vrrp_vr_track_if_t ifs[n_ifs];
            };
        """

        tr_intfs = ast.literal_eval(tr_intfs)
        n_ifs = len(tr_intfs)

        ret = self._vpp.api.vrrp_vr_track_if_add_del(sw_if_index=sw_if_index,
                                                     is_ipv6=is_ipv6,
                                                     vr_id=vr_id,
                                                     is_add=is_add,
                                                     n_ifs=n_ifs,
                                                     addrs=tr_intfs)
        return ret.retval

    def vrrp_vr_get_details(self, vr_id_config, sw_if_index=0):
        if not os.path.isfile('/usr/share/vpp/api/plugins/vrrp.api.json'):
            return 2
        """
            /** \brief VRRP: dump virtual router data
                @param client_index - opaque cookie to identify the sender
                @param context - sender context, to match reply w/ request
                @param sw_if_index - interface to use as filter (0,~0 == "all")
            */
            define vrrp_vr_dump {
              u32 client_index;
              u32 context;
              vl_api_interface_index_t sw_if_index;
            };

            /** \brief VRRP: VR dump response
                @param context - sender context which was passed in the request
                @param conf - configuration parameters for the VR
                @param runtime - runtime state for the VR
            */
            define vrrp_vr_details {
              u32 context;
              vl_api_vrrp_vr_conf_t config;
              vl_api_vrrp_vr_runtime_t runtime;
              u8 n_addrs;
              vl_api_address_t addrs[n_addrs];
            };
        """

        state_dict = {'VRRP_API_VR_STATE_INIT': 'Initializing',
                      'VRRP_API_VR_STATE_BACKUP': 'Backup',
                      'VRRP_API_VR_STATE_MASTER': 'Master',
                      'VRRP_API_VR_STATE_INTF_DOWN': 'Interface down'}

        data = self._vpp.api.vrrp_vr_dump(sw_if_index=sw_if_index)

        for d in data:
            result = []
            sw_if_index = d.config.sw_if_index
            vr_id = d.config.vr_id
            priority = d.config.priority
            ipv4 = [str(a) for a in d.addrs if isinstance(a, ipaddress.IPv4Address)]
            ipv6 = [str(a) for a in d.addrs if isinstance(a, ipaddress.IPv6Address)]
            vr_state = d.runtime.state.name

            state = state_dict[vr_state] if vr_state in state_dict else vr_state

            if vr_id == vr_id_config:
                result.append(
                    {'sw_if_index': sw_if_index, 'vr_id': vr_id, 'priority': priority, 'ipv4': ipv4, 'ipv6': ipv6,
                     'vr_state': state})

                return result
    
    def create_af_xdp_interface(self, hostifname, newname, rxq_num=1, rxq_size=0, txq_size=0, mode=AF_XDP_API_MODE_AUTO, flags=AF_XDP_API_FLAGS_SYSCALL_DEFAULT):
        ret = self._vpp.api.af_xdp_create_v2(
            host_if=hostifname,
            name=newname,
            rxq_num=rxq_num,
            rxq_size=rxq_size,
            txq_size=txq_size,
            mode=mode,
            flags=flags)            

        return (ret.retval, ret.sw_if_index)

    def delete_af_xdp_interface(self, hostifname):
        intfc = self.get_if_by_name(ifname=hostifname)
        if not intfc:
            return (-1)
        ret = self._vpp.api.af_xdp_delete(sw_if_index=intfc['sw_if_index'])
        return ret.retval

